from merkly.utils.crypto import merkle_root, keccak
from merkly.utils.math import is_power_2
from typing import List


class MerkleTree(object):
  """
  # 🌳 Merkle Tree
  - You can passa raw data
  - They will hashed by `keccak-256`
  """
  def __init__(self, leafs: List[str]) -> None:
    if not is_power_2(leafs.__len__()):
      raise Exception("size of leafs should be power of 2")

    self.leafs: List[str] = self.__hash_leafs(leafs)

  def __hash_leafs(self, leafs: List[str]) -> List[str]:
    return list(
      map(keccak, leafs)
    )

  @property
  def root(self):
    """
    # Get a root of merkle tree
    """
    return merkle_root(self.leafs)



from typing import List


def keccak(data: str) -> str:
  """
  # Hash `data: str` using keccak256
  - params `data: str`
  - return `hexadecimal: str`

  ```python
  >>> keccak("merkle")
  "326fe0d8a70ab934a7bf9d1323c6d87ee37bbe70079f82e72203b1e07c0c185c"

  >>> keccak("tree")
  "b2510336c6497719adadc7ade198c988520f3349445f074dc729df0f3c2b12ad"

  >>> keccak("bitcoin")
  "7dee6e1aa550de37364ec77e03e62ea56bf42037b8297280de9d844d88444e4d"

  >>> keccak("ethereum")
  "541111248b45b7a8dc3f5579f630e74cb01456ea6ac067d3f4d793245a255155"
  ```
  """
  from sha3 import keccak_256

  return keccak_256(
    data.encode()
  ).hexdigest()


def slicer(x: list):
  """
  # Slice a `x: list[any]` in sublist of 2 items
  - params `x: list[any]`
  - return `list: list[list[any]]

  ```python
  >>> slicer([1,2,3,4])
  [[1, 2], [3, 4]]

  >>> slicer([i for i in range(10)])
  [[0, 1], [2, 3], [4, 5], [6, 7], [8, 9]]
  ```
  """

  return [
    x[i: i + 2]
    for i in range(0, len(x), 2)
  ]


def merkle_root(x: List[str]) -> str:
  """
  # Merkle Root of `x: list[str]` using keccak256
  - params `x: lsit[str]`
  - return `hexadecimal: list[str]`

  ```python
  >>> merkle_root(["a", "b", "c", "d"])
  ["159b0d5005a27c97537ff0e6d1d0d619be408a5e3f2570816b02dc5a18b74f47"]

  >>> merkle_root(["a", "b"])
  ["63a9f18b64ca5a98ad9dba59259edb0710892614501480a9bed568d98450c151"]
  ```
  """
  from merkly.utils.math import is_power_2

  if not is_power_2(len(x)):
    raise Exception(f"PARÔ, {len(x)}")

  elif len(x) == 1:
    return x

  else:
    return merkle_root([
      keccak(i + j) for i,j in slicer(x)
    ])



def is_power_2(x: int) -> bool:
  """
  # Verify if `x: int` is power of 2
  - params `x: int`
  - return `bool`

  ```python
  assert is_power_2(2) == True
  assert is_power_2(3) == False
  assert is_power_2(16) == True
  assert is_power_2(900) == False
  ```
  """

  a: bool = (x & (x - 1) == 0)
  b: bool = x != 0
  return a and b



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['merkly', 'merkly.utils']

package_data = \
{'': ['*']}

install_requires = \
['pysha3>=1.0.2,<2.0.0']

setup_kwargs = {
    'name': 'merkly',
    'version': '0.4.4',
    'description': '🌳 The simple and easy implementation of Merkle Tree',
    'long_description': '# 🌳 Merkly\n\nThe **simple and easy** implementation of **Python Merkle Tree**\n\n---\n[![CodeQL](https://github.com/olivmath/merkly/actions/workflows/codeql-analysis.yml/badge.svg)](https://github.com/olivmath/merkly/actions/workflows/codeql-analysis.yml)\n[![Test](https://github.com/olivmath/merkly/actions/workflows/test.yml/badge.svg?branch=main)](https://github.com/olivmath/merkly/actions/workflows/test.yml)\n![GitHub last commit](https://img.shields.io/github/last-commit/olivmath/merkly)\n![GitHub commit activity](https://img.shields.io/github/commit-activity/m/olivmath/merkly)\n\n[![PyPI](https://img.shields.io/pypi/v/merkly)](https://pypi.org/project/merkly/)\n[![PyPI - Python Version](https://img.shields.io/pypi/pyversions/merkly)](https://pypi.org/project/merkly/)\n[![PyPI - Downloads](https://img.shields.io/pypi/dm/merkly)](https://pypi.org/project/merkly/)\n![PyPI - License](https://img.shields.io/pypi/l/merkly)\n\n## Table of Contents\n\n- [Credits](#credits)\n- [How to install](#how-to-install)\n- [How to works](#how-to-works)\n- [How to use](#how-to-use)\n- [Contributing](#contributing)\n- [License](#license)\n\n## Credits\n\n[![GitHub Contributors Image](https://contrib.rocks/image?repo=olivmath/merkly)](https://github.com/olivmath/merkly/graphs/contributors)\n\n## How to install\n\n```\npoetry add merkly\n```\n\n```\npip install merkly\n```\n\n## How it works\n\nThis library provides a clean and easy to use implementation of the Merkle Tree with the following features:\n\n- [x] Create Leaf\n- [x] Create Root\n- [ ] Create Proof\n- [ ] Validate Leafs\n\n<!-- ![Merkle Tree](assets/merkle-tree.png) -->\n<img width="1664" alt="merkle-tree" src="https://user-images.githubusercontent.com/50037567/174931499-813e680e-c7e0-4055-8159-93497b4a4ee2.png">\n\n## How to Use\n\nCreate a Merkle Tree\n\n```python\nfrom merkly.mtree import MerkleTree\n\nmtree = MerkleTree(\n  [\'a\', \'b\', \'c\', \'d\']\n)\n\nassert mtree.leafs == [\n  \'3ac225168df54212a25c1c01fd35bebfea408fdac2e31ddd6f80a4bbf9a5f1cb\',\n  \'b5553de315e0edf504d9150af82dafa5c4667fa618ed0a6f19c69b41166c5510\',\n  \'0b42b6393c1f53060fe3ddbfcd7aadcca894465a5a438f69c87d790b2299b9b2\',\n  \'f1918e8562236eb17adc8502332f4c9c82bc14e19bfc0aa10ab674ff75b3d2f3\'\n]\n\nassert mtree.root == [\n  \'115cbb4775ed495f3d954dfa47164359a97762b40059d9502895def16eed609c\'\n]\n```\n\n## Contributing\n\n- Before read a code of conduct: **[CODE_OF_CONDUCT](CODE_OF_CONDUCT.md)**\n- Follow the guide of development: **[CONTRIBUTING](CONTRIBUTING.md)**\n\n## License\n\n[MIT](LICENSE)\n',
    'author': 'Lucas Oliveira',
    'author_email': 'olivmath@protonmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/olivmath/merkly.git',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
